/*
** Module   : Utils.java
** Abstract : General purpose utility methods
**
** Copyright (c) 2004-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 ECF 20041122   @18804 Created initial version, consisting of some
**                           Properties object helper methods.
** 002 NVS 20050223   @19935 Added readPassphrase method.
** 003 ECF 20050225   @19962 Added printChainedErrors method.
** 004 ECF 20050225   @20061 Added findFileInPath method. Generalized and
**                           moved from findFile in uast.AstGenerator.
** 005 GES 20050304   @20196 Added file copying methods.
** 006 GES 20050311   @20293 Added isLaterVersion method.
** 007 ECF 20050329   @20519 Modified printChainedErrors method. Now
**                           prints name of throwable type if no message
**                           is present.
** 008 ECF 20050422   @20844 Added indent parameter to printChainedErrors
**                           method. Indents each message line by the
**                           specified number of spaces.
** 009 GES 20050706   @21667 Added append mode to copyFile().
** 010 ECF 20051103   @23222 Added getLog method. Retrieves an Apache
**                           commons-logging logging object scoped to a
**                           particular package.
** 011 ECF 20051104   @23239 Added describeContext method. Uniquely
**                           describes current context for debugging or
**                           logging.
** 012 ECF 20060112   @23897 Added verifyDirectoryNode method. Verifies
**                           existence and class of a required directory
**                           entry.
** 013 GES 20060122   @24013 Added directory query helpers to manage
**                           the entire process of searching for a
**                           specified value across a pre-defined
**                           hierarchy in the directory.
** 014 ECF 20060123   @24039 Modified describeContext method. Combine
**                           session ID, thread ID, and user ID to create
**                           a unique string.
** 015 GES 20060125   @24073 Added directory node search helper.
** 016 NVS 20060226   @24749 Added generic method search and invocation
**                           service invoke().
** 017 NVS 20060227   @24771 Added noHelp() method which is a placeholder
**                           for a real help application.
** 018 ECF 20060309   @24998 Added getUnqualifiedName() method. Gets the
**                           short name of a class or interface.
** 019 GES 20060609   @27053 Removed fillInStackTrace() from invoke()
**                           since this only destroys the exception's
**                           stack trace (it doesn't affect any chained
**                           cause's stack trace). It is better to have
**                           the original record of where the problem
**                           occurred than to know that we rethrew the
**                           exception from this point.
** 020 GES 20060802   @28296 Created primitive/wrapper conversion helpers
**                           to simplify list and array processing.
** 021 ECF 20060918   @29668 Temporary fix to describeContext(). Handle
**                           invalid thread ID returned from
**                           SecurityManager. A better fix would be to
**                           ensure SecurityManager never returns a null
**                           thread ID.
** 022 GES 20061220   @31807 Removed readPassphrase() and replaced it
**                           with prompt() that is a simpler and safer
**                           alternative.
** 023 NVS 20070112   @31893 Added uniqueId() method and supporting data
**                           in form of a context local work area.
**                           Added methods to work with sets: intersects()
**                           intersection() and difference().
** 024 NVS 20070126   @32010 Added scope param to findDirectoryNodePath()
**                           method. Now the method can do server level
**                           searches (/server/<serverId> and
**                           /server/default paths) and account level
**                           searches as the original implementation.
** 025 NVS 20070404   @32764 Added new method diffIntArrays().
** 026 GES 20070716   @34530 Moved file copying related methods here for
**                           better reuse.
** 027 ECF 20071009   @35600 Added getPortForService() and variants of
**                           other directory convenience methods. The
**                           former looks up the port number for a given
**                           service name. The latter add a 'scope'
**                           parameter to permit looking up values of each
**                           basic data type by account/group or by
**                           server.
** 028 ECF 20071101   @35897 Changed invoke() signature. Now uses var-args
**                           instead of an Object array for parameters, if
**                           any, to be passed to the invoked method.
** 029 GES 20080414   @38023 Added helpers for text processing which
**                           honors the proper charset conversions.
** 030 GES 20080421   @38058 Added another charset helper.
** 031 CA  20081006   @40061 Added integerCollectionToPrimitive.
** 032 CA  20081205   @40820 Added integerMapToPrimitive.
** 033 CA  20090413   @41770 Fixed bug in integerMapToPrimitive - index was
**                           not incremented.
** 034 GES 20090422   @41918 Converted to standard string formatting.
** 035 GES 20090424   @41975 Import change.
** 036 NVS 20090624   @42833 Added new method removeDirectoryNode().
** 037 SIY 20090701   @43013 Added scroll() method. 
**                           Removed integerListToPrimitive() which is a copy
**                           of integerCollectionToPrimitive(). Inferred 
**                           generics in order to improve type checking.
** 038 SIY 20090923   @43987 Directory binding in getDirectoryNodeWorker now is
**                           optional if directory is already bound.  
** 039 ECF 20101004          Added methods to match a file path case-
**                           insensitively and to canonicalize a file system
**                           path.
** 040 SIY 20110112          Changes related to new color handling scheme. Minor
**                           cleanups.
** 041 GES 20110623          Fixed case sensitive filename matching helpers.
** 042 CA  20130822          Added API to check if two sets match.
** 043 HC  20131031          Support methods for Long resource IDs, see issue #2183.
** 044 CA  20140206          Improved the directory node lookup APIs, to allow search using an 
**                           explicit account list and to allow fallback to a per-server value, if
**                           per-account search failed.
** 045 OM  20140706          Fixed matchPathCaseInsensitively for Windows filesystem.
** 046 GES 20141106          Improved some javadocs and added unimplemented feature logging.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;
import java.nio.charset.*;
import java.lang.reflect.*;
import org.apache.commons.logging.*;
import com.goldencode.p2j.*;
import com.goldencode.util.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.ui.LogicalTerminal;

/**
 * Collection of useful, static, utility methods for various tasks.
 */
public final class Utils
{
   /** Define search modes for the directory node lookups. */
   public static enum DirScope
   {
      SERVER,
      ACCOUNTS,
      BOTH
   }

   /** Stores context-local state variables. */
   private static ContextContainer work = new ContextContainer(); 

   /** Charset name to use for encoding or <code>null</code> for default. */
   private static String override = null;
   
   /** Tracks if the static data has been initialized. */
   private static boolean init = false;
   
   /** Logging token that ensures unimplemented feature logging occurs only once. */
   private static Object token = null;
   
   /**
    * Private constructor to prevent instances of this class from being
    * created.
    */
   private Utils()
   {
   }

   /**
    * Scroll given array so element with specified index will be first in
    * array. All elements before it will be placed at the end of the array.
    * <p>
    * NOTE: Scrolling is performed "in place", so input array gets updated!
    * 
    * @param    <T>
    * @param    source
    *           Source array to scroll.
    * @param    first
    *           The index of element which will be first.
    *           
    * @return   reference to the source array.
    */
   public static<T> T[] scroll(T[] source, int first)
   {
       if (first <= 0 || first >= source.length - 1)
           return source;
       
       T[] tmp1 = Arrays.copyOfRange(source, 0, first);
       T[] tmp2 = Arrays.copyOfRange(source, first, source.length);
       
       for (int i = 0; i < source.length; i++)
       {
           source[i] = (i < (source.length - first)) ? tmp2[i] : 
                                                       tmp1[i - (source.length - first)];
       }
       
       return source;
   }
   
   /**
    * Convert the array of wrapper objects into the corresponding array of
    * primitive types.
    *
    * @param    array
    *           The array of wrapper objects.  Must not be <code>null</code>.
    *
    * @return   The corresponding primitive array.
    */
   public static int[] objectToPrimitive(Integer[] array)
   {
      int[] out = new int[array.length];
      
      for (int i = 0; i < array.length; i++)
      {
         out[i] = array[i].intValue();
      }
      
      return out;
   }
   
   /**
    * Convert the array of primitives into the corresponding array of wrapper
    * objects.
    *
    * @param    array
    *           The array of primitive types.  Must not be <code>null</code>.
    *
    * @return   The corresponding wrapper object array.
    */
   public static Integer[] primitiveToObject(int[] array)
   {
      Integer[] out = new Integer[array.length];
      
      for (int i = 0; i < array.length; i++)
      {
         out[i] = new Integer(array[i]);
      }
      
      return out;
   }
   
   /**
    * Convert a collection of wrapper objects into the corresponding array of
    * primitive types.
    *
    * @param    list
    *           The collection of wrapper objects. Must not be 
    *           <code>null</code>.
    *
    * @return   The corresponding primitive array.
    */
   public static int[] integerCollectionToPrimitive(Collection<Integer> list)
   {
      return objectToPrimitive(list.toArray(new Integer[list.size()]));
   }

   /**
    * Convert a map of wrapper objects into the corresponding array of
    * primitive types.
    *
    * @param    map
    *           The map of wrapper objects. Must not be 
    *           <code>null</code>.
    *
    * @return   The corresponding primitive arrays. Array on index 0 will hold
    *           the keys and array on index 1 will hold the values.
    */
   public static int[][] integerMapToPrimitive(Map<Integer, Integer> map)
   {
      int[][] res = new int[2][map.size()];
      int idx = 0;
      for (int key : map.keySet())
      {
         res[0][idx] = key;
         res[1][idx] = map.get(key);
         idx++;
      }
      return res;
   }

   /**
    * Convert an array of primitive types into a list of the corresponding
    * wrapper objects.
    *
    * @param    array
    *           The primitive array.  Must not be <code>null</code>.
    *
    * @return   The list of wrapper objects.
    */
   public static List<Integer> primitiveToIntegerList(int[] array)
   {
      return Arrays.asList(primitiveToObject(array));
   }
   
   /**
    * Convert the array of wrapper objects into the corresponding array of
    * primitive types.
    *
    * @param    array
    *           The array of wrapper objects.  Must not be <code>null</code>.
    *
    * @return   The corresponding primitive array.
    */
   public static long[] objectToPrimitive(Long[] array)
   {
      long[] out = new long[array.length];
      
      for (int i = 0; i < array.length; i++)
      {
         out[i] = array[i].longValue();
      }
      
      return out;
   }
   
   /**
    * Convert the array of primitives into the corresponding array of wrapper
    * objects.
    *
    * @param    array
    *           The array of primitive types.  Must not be <code>null</code>.
    *
    * @return   The corresponding wrapper object array.
    */
   public static Long[] primitiveToObject(long[] array)
   {
      Long[] out = new Long[array.length];
      
      for (int i = 0; i < array.length; i++)
      {
         out[i] = new Long(array[i]);
      }
      
      return out;
   }
   
   /**
    * Convert a collection of wrapper objects into the corresponding array of
    * primitive types.
    *
    * @param    list
    *           The collection of wrapper objects. Must not be 
    *           <code>null</code>.
    *
    * @return   The corresponding primitive array.
    */
   public static long[] longCollectionToPrimitive(Collection<Long> list)
   {
      return objectToPrimitive(list.toArray(new Long[list.size()]));
   }
   
   /**
    * Convert an array of primitive types into a list of the corresponding
    * wrapper objects.
    *
    * @param    array
    *           The primitive array.  Must not be <code>null</code>.
    *
    * @return   The list of wrapper objects.
    */
   public static List<Long> primitiveToLongList(long[] array)
   {
      return Arrays.asList(primitiveToObject(array));
   }
   
   /**
    * Load and return a <code>Properties</code> object from a properties
    * file specified by <code>filename</code>.
    *
    * @param   filename
    *          Name of the properties file.
    *
    * @return  Properties object containing the data read from
    *          <code>filename</code>.
    *
    * @throws  IOException
    *          if any error occurs reading the file.
    */
   public static Properties loadProperties(String filename)
   throws IOException
   {
      Properties props = new Properties();
      FileInputStream fis = null;
      BufferedInputStream bis = null;
      try
      {
         // Read properties from file.
         fis = new FileInputStream(filename);
         bis = new BufferedInputStream(fis);
         props.load(bis);
      }
      finally
      {
         // Clean up streams, regardless of success.
         if (bis != null)
         {
            bis.close();
         }
         else if (fis != null)
         {
            fis.close();
         }
      }
      
      return props;
   }
   
   /**
    * Retrieve the property with the specified <code>key</code> from within
    * the specified properties. Assume <code>key</code> represents a required
    * property; an exception is thrown if it is not present.
    *
    * @param   props
    *          Properties object from which to retrieve required property.
    * @param   key
    *          Key to the required property.
    *          
    * @return  property value.
    *
    * @throws  MissingDataException
    *          if the required property is not present.
    */
   public static String getRequiredProperty(Properties props, String key)
   throws MissingDataException
   {
      String value = props.getProperty(key);
      if (value == null)
      {
         throw new MissingDataException(
            "'" + key + "' is a required configuration property");
      }
      
      return value;
   }

   /**
    * Displays a prompt on <code>stdout</code> and returns the input entered
    * by the user on <code>stdin</code>.  The first new line encountered
    * will end the input reading loop and the text will not include this
    * new line character.
    *
    * @param    txt
    *           The text message to display.
    *
    * @return   The text that was interactively entered or <code>null</code>
    *           if there was no input provided.
    *           
    * @throws IOException 
    */
   public static char[] prompt(String txt) 
   throws IOException 
   {
      System.out.print(txt);
      System.out.flush();
      
      StringBuffer sb = new StringBuffer();
      
      int ch;

      readLoop:
      while (true)
      {
         switch (ch = System.in.read())
         {
            case -1:
            case '\n':
               break readLoop;

            case '\f':
            case '\t':
            case '\r':
               continue readLoop;

            default:
               sb.append((char) ch);
               break;
         }
      }

      if (sb.length() == 0) 
      {
         return null;
      }

      return sb.toString().toCharArray();
   }
   
   /**
    * Extract the error message from a throwable and each of its chained
    * throwables, printing them on consecutive lines using the specified
    * print stream. If a throwable contains no error message, no line is
    * printed for that throwable.
    *
    * @param   thr
    *          Throwable from which to extract message and chained throwables.
    * @param   out
    *          Print stream to which messages will be written.
    * @param   indent
    *          Number of spaces to indent each error message.
    */
   public static void printChainedErrors(Throwable thr,
                                         PrintStream out,
                                         int indent)
   {
      Throwable next = thr;
      do
      {
         String msg = next.getMessage();
         if (msg != null && msg.trim().length() > 0)
         {
            out.println(StringHelper.indent(msg, indent));
         }
         else
         {
            msg = next.getClass().getName();
            out.println(StringHelper.indent(msg, indent));
         }
         next = next.getCause();
      }
      while (next != null);
   }
   
   /**
    * Reports the default character set name. Since the Progress default
    * charset is ISO-8859-1, when the JVM reports that UTF-8 is the default,
    * this method will return the string "ISO-8859-1" instead.
    * <p>
    * The assumption (possibly not always correct) is that UTF-8 is not a
    * valid Progress charset and instead, to provide a proper default, we
    * must override with a compatible charset. Note that if the default
    * charset name is not UTF-8, it is assumed that the charset has been
    * overridden already (at JVM startup).  In that case, the default is
    * returned unchanged.
    *
    * @return   A default charset name.
    */
   public static String getCharsetName()
   {
      Charset def  = Charset.defaultCharset();
      String  cset = getCharsetOverride();
      
      return (cset == null) ? def.name() : cset;
   }
   
   /**
    * Reports if the default character set needs to be replaced. The Progress default charset is
    * ISO-8859-1 and when the JVM reports that UTF-8 is the default, this method will return the
    * string "ISO-8859-1" instead.
    * <p>
    * If the 4GL environment is actually configured for UTF-8, then this approach won't work.
    * If the default codepage name for the JVM is not UTF-8, it is assumed that the codepage
    * has been overriden already (at JVM startup).  In that case, the value reported by the JVM
    * is returned unchanged.
    *
    * @return   A default charset name to honor or <code>null</code> if no override is required.
    */
   public static String getCharsetOverride()
   {
      Charset def  = Charset.defaultCharset();
      String  cset = null;
      
      // this test may not be complete, but it handles the most common
      // case where the extended ASCII range will be "munged" by an
      // improper byte to char conversion
      if (def.name().equals("UTF-8"))
      {
         // this gets called too much to allow it to log since we don't want it cluttering
         // client or conversion logs
         // token = UnimplementedFeature.todo("This code doesn't properly handle UTF-8.", token);
         
         // override with a safe default (matches the Progress default) 
         cset = "ISO-8859-1";
      }
      
      return cset;
   }
   
   /**
    * Convert the given integer to a Unicode character using the character
    * set defined as the default for the environment, honoring any override
    * configured.  See {@link #getCharsetOverride}.
    * <p>
    * The normal ASCII range can be converted to a Unicode character by a
    * simple widening conversion from byte to char. But the extended ASCII
    * range (128 through 255 inclusive) will translate to character values
    * that cannot be achieved via a widening conversion.  Instead the
    * proper character set must be used to get the right result.
    *
    * @param    i
    *           The integer to convert.
    *
    * @return   The Unicode character that represents the given integer in the
    *           specified character set. In case of any error, -1 will be
    *           returned.
    */
   public static char toChar(int i)
   {
      synchronized (Utils.class)
      {
         if (!init)
         {
            override = Utils.getCharsetOverride();
            init     = true;
         }
      }
      
      return toChar(i, override);
   }
   
   /**
    * Convert the given byte to a Unicode character using the given character
    * set (or the default charset if the given charset name is
    * <code>null</code>).
    * <p>
    * The normal ASCII range can be converted to a Unicode character by a
    * simple widening conversion from byte to char. But the extended ASCII
    * range (128 through 255 inclusive) will translate to character values
    * that cannot be achieved via a widening conversion.  Instead the
    * proper character set must be used to get the right result.
    *
    * @param    b
    *           The byte to convert.
    * @param    cset
    *           The charset name to use for conversion or <code>null</code>
    *           if the default charset should be used.
    *
    * @return   The Unicode character that represents the given byte in the
    *           specified character set. In case of any error, -1 will be
    *           returned.
    */
   public static char toChar(byte b, String cset)
   {
      String txt = null;
      char   ch  = (char) -1;
      
      try
      {
         if (cset == null)
         {
            // use the default character set to translate this byte
            // into a character
            txt = new String(new byte[] {b});
         }
         else
         {
            // use a specific character set to convert rather than the
            // default
            txt = new String(new byte[] {b}, cset);
         }
         
         ch = txt.charAt(0);
      }
      
      catch (UnsupportedEncodingException uee)
      {
         // nothing to do
      }
      
      return ch;
   }
   
   /**
    * Convert the given byte to a Unicode character using the given character
    * set (or the default charset if the given charset name is
    * <code>null</code>).
    * <p>
    * <b>Only the least significant byte will be converted, all other data is
    * dropped at this time.</b>
    * <p>
    * The normal ASCII range can be converted to a Unicode character by a
    * simple widening conversion from byte to char. But the extended ASCII
    * range (128 through 255 inclusive) will translate to character values
    * that cannot be achieved via a widening conversion.  Instead the
    * proper character set must be used to get the right result.
    *
    * @param    i
    *           The number to convert. Must be between 0 and 255 inclusive.
    * @param    cset
    *           The charset name to use for conversion or <code>null</code>
    *           if the default charset should be used.
    *
    * @return   The Unicode character that represents the given byte in the
    *           specified character set. In case of any error, -1 will be
    *           returned.
    */
   public static char toChar(int i, String cset)
   {
      if (i > 255 || i < 0)
      {
         // TODO: add DBCS support (i > 255 && i < 65534)
         return (char) -1;
      }
      
      return toChar((byte) i, cset);
   }
   
   /**
    * Convert the given Unicode character to an integer value using the given
    * character set (or the default charset if the given charset name is
    * <code>null</code>).
    * <p>
    * The normal ASCII range can be converted from a Unicode character by a
    * simple narrowing conversion from char to byte. But the Unicode
    * characters representing the extended ASCII range (128 through 255
    * inclusive) are not represented in Unicode via a single byte. So a
    * translation from character values to a number cannot be achieved via a
    * narrowing conversion.  Instead the proper character set must be used to
    * get the right result.
    *
    * @param    ch
    *           The Unicode character to convert.
    * @param    cset
    *           The charset name to use for conversion or <code>null</code>
    *           if the default charset should be used.
    *
    * @return   The integer value that represents the given the Unicode 
    *           character in the specified character set or -1 on any error.
    */
   public static int toInt(char ch, String cset)
   {
      int i = -1;
      
      try
      {
         String encoded = new String(new char[] {ch});
         byte[] b = (cset == null) ? encoded.getBytes()
                                   : encoded.getBytes(cset);
         i = b[0] & 0x000000FF;
      }
      
      catch (UnsupportedEncodingException uee)
      {
         // let -1 be returned
      }
      
      return i;
   }
   
   /**
    * Creates all subdirectories needed to provide a valid path for the
    * target filename that is passed in.  It is expected that there will 
    * always be a path included in this filename such that making 
    * directories makes sense.
    *
    * @param    tar
    *           Target filename for which paths must be created.
    *
    * @return   <code>true</code> if the full necessary path already
    *           exists or if it has been created, <code>false</code>
    *           otherwise.
    */
   public static boolean createPathForTarget(String tar)
   {
      String path = tar.substring(0, tar.lastIndexOf(File.separatorChar));
      File   file = new File(path);
      
      // anything to do?
      if (file.exists() && file.isDirectory())
         return true;
      
      try
      {
         // best effort to create the directories including all needed
         // parent dirs
         file.mkdirs();
      }
      catch (Exception exc)
      {
         // some failure, probably a security violation
         return false;
      }
      
      // it better be there now...
      if (file.exists() && file.isDirectory())
         return true;

      // some other unexpected problem!
      return false;
   }
   
   /**
    * Find the specified file in the file system using the specified search
    * paths and return it. Handles absolute and relative paths.
    *
    * @param   filename
    *          Name of the source file to find; if this is not an absolute
    *          path, it should be relative to one of the paths in the
    *          <code>searchPaths</code> array.
    * @param   paths
    *          Array of absolute paths which are prepended to a relative
    *          <code>filename</code>.
    *
    * @return  The file if it was found.
    *
    * @throws  FileNotFoundException
    *          if a file named <code>filename</code> does not exist in the
    *          specified search paths (or at the absolute path provided).
    */
   public static File findFileInPath(String filename, String[] paths)
   throws FileNotFoundException
   {
      File file = new File(filename);
      
      if (file.isAbsolute())
      {
         // Return file if it is absolute and exists.
         if (file.exists())
         {
            return file;
         }
      }
      else
      {
         // Use search paths to resolve relative references.
         int len = paths.length;
         for (int i = 0; i < len; i++)
         {
            file = new File(paths[i], filename);
            if (file.exists())
            {
               return file;
            }
         }
      }
      
      // Reassemble path for error message.
      StringBuffer buf = new StringBuffer();
      String sep = File.pathSeparator;
      int len = paths.length;
      for (int i = 0; i < len; i++)
      {
         buf.append(paths[i])
            .append(sep);
      }
      
      throw new FileNotFoundException(
         "File '" + filename + "' not found in path: " + buf);
   }
   
   /**
    * Canonicalize the given file system path as described by
    * <code>java.io.File.getCanonicalPath()</code>.  In addition, normalize
    * the use of the file separator by replacing all instances of
    * <code>oldSeparator</code> with <code>newSeparator</code>.
    * 
    * @param   path
    *          A relative or absolute path to be canonicalized.
    * @param   oldSeparator
    *          File separator to be replaced.
    * @param   newSeparator
    *          New file separator.
    *
    * @return  Canonicalized path containing normalized file separators.
    * 
    * @throws  IOException
    *          If there is a file system while canonicalizing the path.
    */
   public static String canonicalizePath(String path,
                                         String oldSeparator,
                                         String newSeparator)
   throws IOException
   {
      String res = (new File(path)).getCanonicalPath();
      if (!oldSeparator.equals(newSeparator))
      {
         if ("\\".equals(oldSeparator))
         {
            oldSeparator = "\\\\";
         }
         if ("\\".equals(newSeparator))
         {
            newSeparator = "\\\\";
         }
         res = res.replaceAll(oldSeparator, newSeparator);
      }
      
      return res;
   }
      
   /**
    * Determines if a both source and target exist as files and if the target
    * has a later timestamp.  It does not consider contents or size in this
    * determination.
    *
    * @param    source
    *           Source filename.
    * @param    target
    *           Target filename.
    *
    * @return   <code>true</code> if the both files exist and the target is
    *           newer, otherwise <code>false</code> (includes failures).
    */
   public static boolean isLaterVersion(String source, String target)
   {      
      File src = new File(source);
      File tar = new File(target);
      
      // make sure the source file is OK
      boolean srcOK = src.exists() && src.isFile();
      
      if (!srcOK)
         return false;
      
      // make sure the target file is not there or if it is there that
      // it is older than the source
      if (tar.exists())
      {
         if (tar.isFile())
         {
            // returns true if the source is older, false if the target is
            // older
            return src.lastModified() < tar.lastModified();
         }
      }
      
      return false;
   }     
   
   /**
    * Creates the directory tree to the target if needed and then copies
    * the source to the target if needed.  The copy is byte for byte, no
    * modifications of the file will occur.
    *
    * @param    src
    *           Source filename.
    * @param    tar
    *           Target filename.
    *
    * @return   <code>true</code> if the file was copied successfully OR
    *           if the file didn't need to be copied, <code>false</code>
    *           on any failure.
    */
   public static boolean simpleCopy(String src, String tar)
   {
      // try to create the directory
      if (!createPathForTarget(tar))
      {
         // dir creation failed
         return false;
      }
      
      return copyFileIfNeeded(src, tar, 0);
   }
   
   /**
    * Conditionally copies a file, byte for byte (no changes) to a new file,
    * if and only if the source is an accessible file and the target doesn't
    * exist or the target exists as a file but is older than the source or
    * has a different size.
    *
    * @param    source
    *           Source filename.
    * @param    target
    *           Target filename.
    * @param    bsize
    *           Buffer size or if 0 the buffer size will default to the file
    *           size or 64Kb whichever is less.
    *
    * @return   <code>true</code> if the file was copied successfully OR if
    *           the file didn't need to be copied, <code>false</code> on any
    *           failure.
    */
   public static boolean copyFileIfNeeded(String source,
                                          String target,
                                          int    bsize)
   {      
      File src = new File(source);
      File tar = new File(target);
      
      // make sure the source file is OK
      boolean srcOK = src.exists() && src.isFile();
      
      if (!srcOK)
         return false;
      
      boolean tarOK = true; 
      
      // make sure the target file is not there or if it is there that
      // it is older than the source or it is not the same size, before
      // trying to copy
      if (tar.exists())
      {
         if (tar.isFile())
         {
            boolean oldSrc = src.lastModified() > tar.lastModified();
            long    diff   = src.length() - tar.length();               
            tarOK = oldSrc || (diff != 0);
         }
         else
         {
            return false;
         }
      }
      
      boolean result = false;
      
      // already checked src, now check target
      if (tarOK)
      {
         // best efforts attempt to copy
         result = copyFile(source, target, 0, false);
      }
      
      return result;
   }   
   
   /**
    * Handles the byte for byte copying of a single source file to a target 
    * file, using a buffer size of the caller's choice. The resulting target
    * file will be a newly created file (date/timestamps will be new).
    *
    * @param    source
    *           Source filename.
    * @param    target
    *           Target filename.
    * @param    bsize
    *           Buffer size or if 0 the buffer size will default to the file
    *           size or 64Kb whichever is less.
    * @param    append
    *           <code>true</code> if the output file should be opened in
                append mode.
    *
    * @return   <code>true</code> if the file was copied successfully,
    *           <code>false</code> on any failure.
    */
   public static boolean copyFile(String  source,
                                  String  target,
                                  int     bsize,
                                  boolean append)                                  
   {
      File in  = new File(source);
      long len = in.length();
      
      // if an invalid buffer size is set, the lesser of 64Kb or the file
      // length is used
      if (bsize < 1)
      {
         bsize = 64*1024;
         
         if (len < bsize)
            bsize = (int) len;
      }
      
      try
      {
         FileInputStream  src = new FileInputStream(in);
         FileOutputStream tar = new FileOutputStream(target, append);
         
         int off = 0;
         
         byte[] buf = new byte[bsize];
   
         while (len > bsize)
         {
            src.read(buf, off, bsize);
            tar.write(buf, 0, bsize);
            len -= bsize;
         }
         
         src.read(buf, off, (int)len);
         tar.write(buf, 0, (int)len);
         
         src.close();
         tar.close();
      }
      catch (Exception exc)
      {
         return false;
      }
      
      return true;
   }
   
   /**
    * Supply a list of files which case-insensitively match the given file
    * name.  On a case-insensitive file system, this list will contain no more
    * than one match.  However, on a case-sensitive file system, the list may
    * contain more matches.
    * 
    * @param   path
    *          The file name to be matched.  It will be canonicalized before
    *          the search is conducted.
    * @param   separator
    *          Legacy system file separator which will be replaced with the
    *          target system's file separator (in the result list).
    *          
    * @return  A list containing 0 or more files.  This method will never
    *          return <code>null</code>.
    *          
    * @throws  IOException
    *          if there is any error canonicalizing the <code>path</code>.
    */
   public static List<File>
   getCaseInsensitiveFilenameMatches(String path, String separator)
   throws IOException
   {
      // canonicalize path and ensure file separators are appropriate for the
      // current platform's file system
      String canonicalPath = canonicalizePath(path,
                                              separator,
                                              File.separator);
      
      return getCaseInsensitiveFilenameMatches(canonicalPath);
   }
   
   /**
    * Supply a list of files which case-insensitively match the given file
    * name.  On a case-insensitive file system, this list will contain no more
    * than one match.  However, on a case-sensitive file system, the list may
    * contain more matches.
    * 
    * @param   path
    *          The file name to be matched.  It must already be canonicalized
    *          (see {@link #canonicalizePath}) and with all file separators
    *          converted to the current system's file separator character.
    *          
    * @return  A list containing 0 or more files.  This method will never
    *          return <code>null</code>.
    *          
    * @throws  IOException
    *          if there is any error canonicalizing the <code>path</code>.
    */
   public static List<File> getCaseInsensitiveFilenameMatches(String path)
   throws IOException
   {
      List<File> list = new ArrayList<File>(1);
      matchPathCaseInsensitively(null, path, list);
      
      return list;
   }
   
   /**
    * Retrieve a logging object from Apache commons <code>LogFactory</code>.
    * The logger retrieved is scoped to the package name of the specified
    * class;  thus, classes in the same package which use this method to
    * retrieve loggers will be sharing the same <code>Log</code> instance.
    *
    * @param   clazz
    *          Class object whose package name is used to scope the Log.
    *
    * @return  Logging object for the desired package.
    */
   public static Log getLog(Class<?> clazz)
   {
      String name = clazz.getName();
      name = name.substring(0, name.lastIndexOf("."));
      
      return LogFactory.getLog(name);
   }
   
   /**
    * Get a string which uniquely describes the current context.  This is
    * intended primarily for debugging or logging purposes.
    *
    * @return  Description of context.
    */
   public static String describeContext()
   {
      SecurityManager security = SecurityManager.getInstance();
      if (security == null)
      {
         return Thread.currentThread().getName();
      }
      
      StringBuffer buf = new StringBuffer();
      
      // Session ID.
      buf.append(String.format("%08X", security.getSessionId()));
      buf.append(":");
      
      // Thread ID.
      Object tid = security.getThreadId();
      if (tid != null)
      {
         buf.append(String.format("%08X", tid));
      }
      else
      {
         buf.append("????????");
      }
      buf.append(":");
      
      // User ID.
      buf.append(security.getUserId());
      
      return buf.toString();
   }
   
   /**
    * Executes the given entry point using reflection, instantiating
    * the target class first, if necessary.
    * Catches the exceptions and rethrows those of them that represent
    * Progress conditions.
    *
    * @param    cl
    *           The class to be instantiated. Must be public. Must have a
    *           public, default constructor, unless <code>entry</code> is a
    *           static method.
    * @param    entry
    *           The method to be executed. Must exist within <code>cl</code>.
    *           May be an instance method or static method;  must be
    *           publicly accessible.
    * @param    args
    *           If the method's signature contains parameters, this is a
    *           variable argument list containing those parameters.
    *           
    * @return   The result of the invocation.  If the signature of the method
    *           specifies a <code>void</code> return then this will return
    *           <code>null</code>.  Note that it is a limitation of the
    *           design of Java method signatures such that we do not
    *           distinguish between a method with a <code>void</code> return
    *           and a method that has a real return value which is set to
    *           <code>null</code>.
    * @throws NullPointerException 
    * @throws IllegalAccessException 
    * @throws IllegalArgumentException 
    * @throws InvocationTargetException 
    * @throws InstantiationException 
    */
   public static Object invoke(Class<?> cl, Method entry, Object... args)
   throws NullPointerException,
          IllegalAccessException,
          IllegalArgumentException,
          InvocationTargetException,
          InstantiationException
   {
      // simple test for a quick out
      if (entry == null)
      {
         throw new NullPointerException("Specified entry point is null!");
      }

      Object result = null;
      
      try
      {
         Object instance = null;
         
         // instantiate the class (for non-static method)
         if ((entry.getModifiers() & Modifier.STATIC) == 0)
         {
            instance = cl.newInstance();
         }
         
         // invoke the method
         result = entry.invoke(instance, args);
      }
      catch (InvocationTargetException exc)
      {
         // if the exception thrown was a P2J runtime type, we
         // must handle it normally, so extract the root cause
         // exception and rethrow it
         Throwable cause = exc.getCause();
         
         if (cause != null)
         {
            if (cause instanceof ConditionException)
            {
               throw (ConditionException) cause;
            }
            if (cause instanceof RetryUnwindException)
            {
               throw (RetryUnwindException) cause;
            }
            
            // exception is a runtime problem that we don't handle here
            throw exc;
         }
      }
      return result;
   }

   /**
    * Verify that a directory node at a particular, required directory path
    * exists, and that the class of the node found at that path is that which
    * is expected.
    *
    * @param   ds
    *          Directory service.
    * @param   id
    *          ID of node to verify.
    * @param   expected
    *          Expected class of node.
    *
    * @throws  RuntimeException
    *          if directory node does not exist or is of the incorrect class.
    */
   public static void verifyDirectoryNode(DirectoryService ds,
                                          String id,
                                          String expected)
   throws RuntimeException
   {
      String nodeClass = ds.getNodeClass(id);
      
      if (nodeClass == null)
      {
         throw new RuntimeException(
            "Directory entry missing:  " + id);
      }
      
      if (!nodeClass.equals(expected))
      {
         throw new RuntimeException(
            "Directory entry is incorrect type:  " + id);
      }
   }
   
   /**
    * Obtain the specified attribute from the directory if it exists or return
    * the default value if the directory node doesn't exist with the proper
    * type, the directory paths don't exist or the directory is unavailable. 
    * <p>
    * This method implements a search algorithm that allows values to be
    * found that are account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/&lt;id&gt;
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/&lt;id&gt;
    * <p>
    * No exceptions are thrown, the given initial value will always be safely
    * returned.
    *
    * @param    ds
    *           A bound directory service instance.  If this is 
    *           <code>null</code> then the instance will be dynamically
    *           queried and bound for the duration of the method and then
    *           unbound at the end.  This is a convenience for situations
    *           in which only a single value is needed.  This has negative
    *           performance implications in the case where multiple values
    *           are sequentially obtained.
    * @param    nodeId
    *           ID of node to obtain.  This will be appended to the various
    *           paths listed above (and in the precedence order specified 
    *           above).
    * @param    initial
    *           The default value to return if the directory backed value
    *           cannot be obtained for ANY reason.
    *
    * @return   The requested value or the given initial value if for ANY
    *           reason the directory backed value cannot be obtained.
    */
   public static int getDirectoryNodeInt(DirectoryService ds,
                                         String           nodeId,
                                         int              initial)
   {
      return getDirectoryNodeInt(ds, nodeId, initial, DirScope.ACCOUNTS, null);
   }
      
   /**
    * Obtain the specified attribute from the directory if it exists or return
    * the default value if the directory node doesn't exist with the proper
    * type, the directory paths don't exist or the directory is unavailable. 
    * <p>
    * This method implements a search algorithm that allows values to be
    * found that are account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/&lt;id&gt;
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/&lt;id&gt;
    * <p>
    * No exceptions are thrown, the given initial value will always be safely
    * returned.
    *
    * @param    ds
    *           A bound directory service instance.  If this is 
    *           <code>null</code> then the instance will be dynamically
    *           queried and bound for the duration of the method and then
    *           unbound at the end.  This is a convenience for situations
    *           in which only a single value is needed.  This has negative
    *           performance implications in the case where multiple values
    *           are sequentially obtained.
    * @param    nodeId
    *           ID of node to obtain.  This will be appended to the various
    *           paths listed above (and in the precedence order specified 
    *           above).
    * @param    initial
    *           The default value to return if the directory backed value
    *           cannot be obtained for ANY reason.
    *
    * @return   The requested value or the given initial value if for ANY
    *           reason the directory backed value cannot be obtained.
    */
   public static char getDirectoryNodeChar(DirectoryService ds,
                                           String           nodeId,
                                           char             initial)
   {
      return getDirectoryNodeChar(ds, nodeId, initial, DirScope.ACCOUNTS, null);
   }
   
   /**
    * Obtain the specified attribute from the directory if it exists or return
    * the default value if the directory node doesn't exist with the proper
    * type, the directory paths don't exist or the directory is unavailable. 
    * <p>
    * This method implements a search algorithm that allows values to be
    * found that are account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/&lt;id&gt;
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/&lt;id&gt;
    * <p>
    * No exceptions are thrown, the given initial value will always be safely
    * returned.
    *
    * @param    ds
    *           A bound directory service instance.  If this is 
    *           <code>null</code> then the instance will be dynamically
    *           queried and bound for the duration of the method and then
    *           unbound at the end.  This is a convenience for situations
    *           in which only a single value is needed.  This has negative
    *           performance implications in the case where multiple values
    *           are sequentially obtained.
    * @param    nodeId
    *           ID of node to obtain.  This will be appended to the various
    *           paths listed above (and in the precedence order specified 
    *           above).
    * @param    initial
    *           The default value to return if the directory backed value
    *           cannot be obtained for ANY reason.
    *
    * @return   The requested value or the given initial value if for ANY
    *           reason the directory backed value cannot be obtained.
    */
   public static String getDirectoryNodeString(DirectoryService ds,
                                               String           nodeId,
                                               String           initial)
   {
      return getDirectoryNodeString(ds, nodeId, initial, DirScope.ACCOUNTS, null);
   }
      
   /**
    * Obtain the specified attribute from the directory if it exists or return
    * the default value if the directory node doesn't exist with the proper
    * type, the directory paths don't exist or the directory is unavailable. 
    * <p>
    * This method implements a search algorithm that allows values to be
    * found that are account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/&lt;id&gt;
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/&lt;id&gt;
    * <p>
    * No exceptions are thrown, the given initial value will always be safely
    * returned.
    *
    * @param    ds
    *           A bound directory service instance.  If this is 
    *           <code>null</code> then the instance will be dynamically
    *           queried and bound for the duration of the method and then
    *           unbound at the end.  This is a convenience for situations
    *           in which only a single value is needed.  This has negative
    *           performance implications in the case where multiple values
    *           are sequentially obtained.
    * @param    nodeId
    *           ID of node to obtain.  This will be appended to the various
    *           paths listed above (and in the precedence order specified 
    *           above).
    * @param    initial
    *           The default value to return if the directory backed value
    *           cannot be obtained for ANY reason.
    *
    * @return   The requested value or the given initial value if for ANY
    *           reason the directory backed value cannot be obtained.
    */
   public static boolean getDirectoryNodeBoolean(DirectoryService ds,
                                                 String           nodeId,
                                                 boolean          initial)
   {
      return getDirectoryNodeBoolean(ds, nodeId, initial, DirScope.ACCOUNTS, null);
   }
   
   /**
    * Obtain the specified attribute from the directory if it exists or return
    * the default value if the directory node doesn't exist with the proper
    * type, the directory paths don't exist or the directory is unavailable. 
    * <p>
    * This method implements a search algorithm that allows values to be
    * found that are account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/&lt;id&gt;
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/&lt;id&gt;
    * <p>
    * No exceptions are thrown, the given initial value will always be safely
    * returned.
    *
    * @param    ds
    *           A bound directory service instance.  If this is 
    *           <code>null</code> then the instance will be dynamically
    *           queried and bound for the duration of the method and then
    *           unbound at the end.  This is a convenience for situations
    *           in which only a single value is needed.  This has negative
    *           performance implications in the case where multiple values
    *           are sequentially obtained.
    * @param    nodeId
    *           ID of node to obtain.  This will be appended to the various
    *           paths listed above (and in the precedence order specified 
    *           above).
    * @param    initial
    *           The default value to return if the directory backed value
    *           cannot be obtained for ANY reason.
    * @param    scope 
    *           <code>true</code> if account related search, otherwise server
    *           related search is assumed.
    *           
    * @return   The requested value or the given initial value if for ANY
    *           reason the directory backed value cannot be obtained.
    */
   public static int getDirectoryNodeInt(DirectoryService ds,
                                         String           nodeId,
                                         int              initial,
                                         boolean          scope)
   {
      return getDirectoryNodeInt(ds, nodeId, initial,
                                 (scope ? DirScope.ACCOUNTS : DirScope.SERVER), null);
   }
   
   /**
    * Obtain the specified attribute from the directory if it exists or return
    * the default value if the directory node doesn't exist with the proper
    * type, the directory paths don't exist or the directory is unavailable. 
    * <p>
    * This method implements a search algorithm that allows values to be
    * found that are account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/&lt;id&gt;
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/&lt;id&gt;
    * <p>
    * No exceptions are thrown, the given initial value will always be safely
    * returned.
    *
    * @param    ds
    *           A bound directory service instance.  If this is 
    *           <code>null</code> then the instance will be dynamically
    *           queried and bound for the duration of the method and then
    *           unbound at the end.  This is a convenience for situations
    *           in which only a single value is needed.  This has negative
    *           performance implications in the case where multiple values
    *           are sequentially obtained.
    * @param    nodeId
    *           ID of node to obtain.  This will be appended to the various
    *           paths listed above (and in the precedence order specified 
    *           above).
    * @param    initial
    *           The default value to return if the directory backed value
    *           cannot be obtained for ANY reason.
    * @param    scope
    *           <code>true</code> if account related search, otherwise server
    *           related search is assumed.
    *
    * @return   The requested value or the given initial value if for ANY
    *           reason the directory backed value cannot be obtained.
    */
   public static char getDirectoryNodeChar(DirectoryService ds,
                                           String           nodeId,
                                           char             initial,
                                           boolean          scope)
   {
      return getDirectoryNodeChar(ds, nodeId, initial,
                                  (scope ? DirScope.ACCOUNTS : DirScope.SERVER), null);
   }

   /**
    * Obtain the specified attribute from the directory if it exists or return
    * the default value if the directory node doesn't exist with the proper
    * type, the directory paths don't exist or the directory is unavailable. 
    * <p>
    * This method implements a search algorithm that allows values to be
    * found that are account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/&lt;id&gt;
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/&lt;id&gt;
    * <p>
    * No exceptions are thrown, the given initial value will always be safely
    * returned.
    *
    * @param    ds
    *           A bound directory service instance.  If this is 
    *           <code>null</code> then the instance will be dynamically
    *           queried and bound for the duration of the method and then
    *           unbound at the end.  This is a convenience for situations
    *           in which only a single value is needed.  This has negative
    *           performance implications in the case where multiple values
    *           are sequentially obtained.
    * @param    nodeId
    *           ID of node to obtain.  This will be appended to the various
    *           paths listed above (and in the precedence order specified 
    *           above).
    * @param    initial
    *           The default value to return if the directory backed value
    *           cannot be obtained for ANY reason.
    * @param    scope
    *           <code>true</code> if account related search, otherwise server
    *           related search is assumed.
    *
    * @return   The requested value or the given initial value if for ANY
    *           reason the directory backed value cannot be obtained.
    */
   public static String getDirectoryNodeString(DirectoryService ds,
                                               String           nodeId,
                                               String           initial,
                                               boolean          scope)
   {
      return getDirectoryNodeString(ds, nodeId, initial,
                                    (scope ? DirScope.ACCOUNTS : DirScope.SERVER), null);
   }
   
   /**
    * Obtain the specified attribute from the directory if it exists or return
    * the default value if the directory node doesn't exist with the proper
    * type, the directory paths don't exist or the directory is unavailable. 
    * <p>
    * This method implements a search algorithm that allows values to be
    * found that are account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/&lt;id&gt;
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/&lt;id&gt;
    * <p>
    * No exceptions are thrown, the given initial value will always be safely
    * returned.
    *
    * @param    ds
    *           A bound directory service instance.  If this is 
    *           <code>null</code> then the instance will be dynamically
    *           queried and bound for the duration of the method and then
    *           unbound at the end.  This is a convenience for situations
    *           in which only a single value is needed.  This has negative
    *           performance implications in the case where multiple values
    *           are sequentially obtained.
    * @param    nodeId
    *           ID of node to obtain.  This will be appended to the various
    *           paths listed above (and in the precedence order specified 
    *           above).
    * @param    initial
    *           The default value to return if the directory backed value
    *           cannot be obtained for ANY reason.
    * @param    scope
    *           <code>true</code> if account related search, otherwise server
    *           related search is assumed.
    *
    * @return   The requested value or the given initial value if for ANY
    *           reason the directory backed value cannot be obtained.
    */
   public static boolean getDirectoryNodeBoolean(DirectoryService ds,
                                                 String           nodeId,
                                                 boolean          initial,
                                                 boolean          scope)
   {
      return getDirectoryNodeBoolean(ds, nodeId, initial,
                                     (scope ? DirScope.ACCOUNTS : DirScope.SERVER), null);
   }
   
   /**
    * Obtain the path to the node in the directory if it exists or return
    * <code>null</code> if the directory node doesn't exist with the proper
    * type, the directory paths don't exist or the directory is unavailable. 
    * <p>
    * This method implements a search algorithm that allows paths to be
    * found that are account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * There are two search scopes: account relayed and server related.
    * Account related search is described below. The server related search
    * is a version of the search that is limited to the first two server
    * related levels.
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/&lt;id&gt;
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/&lt;id&gt;
    *
    * @param    ds
    *           A bound directory service instance.  If this is 
    *           <code>null</code> then the instance will be dynamically
    *           queried and bound for the duration of the method and then
    *           unbound at the end.  This is a convenience for situations
    *           in which only a single value is needed.  This has negative
    *           performance implications in the case where multiple values
    *           are sequentially obtained.
    * @param    nodeId
    *           ID of node to obtain.  This will be appended to the various
    *           paths listed above (and in the precedence order specified 
    *           above).
    * @param    expected
    *           Expected class of node.  This will be checked with that
    *           found in the directory for consistency.
    * @param    scope
    *           <code>true</code> if account related search, otherwise server
    *           related search is assumed.
    *
    * @return   The path to the found node or <code>null</code> if for ANY
    *           reason the directory node cannot be found.
    */
   public static String findDirectoryNodePath(DirectoryService ds,
                                              String           nodeId,
                                              String           expected,
                                              boolean          scope)
   {
      return findDirectoryNodePath(ds, nodeId, expected,
                                   (scope ? DirScope.ACCOUNTS : DirScope.SERVER), null);
   }

   /**
    * Obtain the specified attribute from the directory if it exists or return
    * the default value if the directory node doesn't exist with the proper
    * type, the directory paths don't exist or the directory is unavailable. 
    * <p>
    * This method implements a search algorithm that allows values to be
    * found that are account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/&lt;id&gt;
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/&lt;id&gt;
    * <p>
    * No exceptions are thrown, the given initial value will always be safely
    * returned.
    * <p>
    * If the {@code scope} is set to {@link DirScope#BOTH}, first a per-account search will be
    * performed.  If the value is not found, it fallbacks to a per-server search.
    *
    * @param    ds
    *           A bound directory service instance.  If this is 
    *           <code>null</code> then the instance will be dynamically
    *           queried and bound for the duration of the method and then
    *           unbound at the end.  This is a convenience for situations
    *           in which only a single value is needed.  This has negative
    *           performance implications in the case where multiple values
    *           are sequentially obtained.
    * @param    nodeId
    *           ID of node to obtain.  This will be appended to the various
    *           paths listed above (and in the precedence order specified 
    *           above).
    * @param    initial
    *           The default value to return if the directory backed value
    *           cannot be obtained for ANY reason.
    * @param    scope 
    *           Search mode for the directory node lookup.
    * @param    accountIds
    *           Explicit list of account IDs which should be included in the search. When
    *           {@code null}, the current context's account IDs are resolved and used.
    *           
    * @return   The requested value or the given initial value if for ANY
    *           reason the directory backed value cannot be obtained.
    */
   public static int getDirectoryNodeInt(DirectoryService ds,
                                         String           nodeId,
                                         int              initial,
                                         DirScope         scope,
                                         String[]         accountIds)
   {
      DirectoryHelper helper = new DirectoryHelper("value")
      {
         @Override
         public Object getDirectoryNode(DirectoryService ds, String nodeId)
         {
            return ds.getNodeInteger(nodeId, attrName);
         }
      };
      
      Integer result = (Integer) getDirectoryNodeWorker(ds,
                                                        nodeId,
                                                        "integer",
                                                        helper,
                                                        scope,
                                                        accountIds);
      
      return (result == null) ? initial : result.intValue();
   }
      
   /**
    * Obtain the specified attribute from the directory if it exists or return
    * the default value if the directory node doesn't exist with the proper
    * type, the directory paths don't exist or the directory is unavailable. 
    * <p>
    * This method implements a search algorithm that allows values to be
    * found that are account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/&lt;id&gt;
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/&lt;id&gt;
    * <p>
    * No exceptions are thrown, the given initial value will always be safely
    * returned.
    * <p>
    * If the {@code scope} is set to {@link DirScope#BOTH}, first a per-account search will be
    * performed.  If the value is not found, it fallbacks to a per-server search.
    *
    * @param    ds
    *           A bound directory service instance.  If this is 
    *           <code>null</code> then the instance will be dynamically
    *           queried and bound for the duration of the method and then
    *           unbound at the end.  This is a convenience for situations
    *           in which only a single value is needed.  This has negative
    *           performance implications in the case where multiple values
    *           are sequentially obtained.
    * @param    nodeId
    *           ID of node to obtain.  This will be appended to the various
    *           paths listed above (and in the precedence order specified 
    *           above).
    * @param    initial
    *           The default value to return if the directory backed value
    *           cannot be obtained for ANY reason.
    * @param    scope 
    *           Search mode for the directory node lookup.
    * @param    accountIds
    *           Explicit list of account IDs which should be included in the search. When
    *           {@code null}, the current context's account IDs are resolved and used.
    *
    * @return   The requested value or the given initial value if for ANY
    *           reason the directory backed value cannot be obtained.
    */
   public static char getDirectoryNodeChar(DirectoryService ds,
                                           String           nodeId,
                                           char             initial,
                                           DirScope         scope,
                                           String[]         accountIds)
   {
      DirectoryHelper helper = new DirectoryHelper("value")
      {
         @Override
         public Object getDirectoryNode(DirectoryService ds, String nodeId)
         {
            return ds.getNodeString(nodeId, attrName);
         }
      };
      
      String result = (String) getDirectoryNodeWorker(ds,
                                                      nodeId,
                                                      "string",
                                                      helper,
                                                      scope,
                                                      accountIds);
      
      return (result != null && result.length() > 0) ? result.charAt(0)
                                                     : initial;
   }
   
   /**
    * Obtain the specified attribute from the directory if it exists or return
    * the default value if the directory node doesn't exist with the proper
    * type, the directory paths don't exist or the directory is unavailable. 
    * <p>
    * This method implements a search algorithm that allows values to be
    * found that are account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/&lt;id&gt;
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/&lt;id&gt;
    * <p>
    * No exceptions are thrown, the given initial value will always be safely
    * returned.
    * <p>
    * If the {@code scope} is set to {@link DirScope#BOTH}, first a per-account search will be
    * performed.  If the value is not found, it fallbacks to a per-server search.
    *
    * @param    ds
    *           A bound directory service instance.  If this is 
    *           <code>null</code> then the instance will be dynamically
    *           queried and bound for the duration of the method and then
    *           unbound at the end.  This is a convenience for situations
    *           in which only a single value is needed.  This has negative
    *           performance implications in the case where multiple values
    *           are sequentially obtained.
    * @param    nodeId
    *           ID of node to obtain.  This will be appended to the various
    *           paths listed above (and in the precedence order specified 
    *           above).
    * @param    initial
    *           The default value to return if the directory backed value
    *           cannot be obtained for ANY reason.
    * @param    scope 
    *           Search mode for the directory node lookup.
    * @param    accountIds
    *           Explicit list of account IDs which should be included in the search. When
    *           {@code null}, the current context's account IDs are resolved and used.
    *
    * @return   The requested value or the given initial value if for ANY
    *           reason the directory backed value cannot be obtained.
    */
   public static String getDirectoryNodeString(DirectoryService ds,
                                               String           nodeId,
                                               String           initial,
                                               DirScope         scope,
                                               String[]         accountIds)
   {
      DirectoryHelper helper = new DirectoryHelper("value")
      {
         @Override
         public Object getDirectoryNode(DirectoryService ds, String nodeId)
         {
            return ds.getNodeString(nodeId, attrName);
         }
      };
      
      String result = (String) getDirectoryNodeWorker(ds,
                                                      nodeId,
                                                      "string",
                                                      helper,
                                                      scope,
                                                      accountIds);
      
      return (result == null) ? initial : result;
   }
      
   /**
    * Obtain the specified attribute from the directory if it exists or return
    * the default value if the directory node doesn't exist with the proper
    * type, the directory paths don't exist or the directory is unavailable. 
    * <p>
    * This method implements a search algorithm that allows values to be
    * found that are account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/&lt;id&gt;
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/&lt;id&gt;
    * <p>
    * No exceptions are thrown, the given initial value will always be safely
    * returned.
    * <p>
    * If the {@code scope} is set to {@link DirScope#BOTH}, first a per-account search will be
    * performed.  If the value is not found, it fallbacks to a per-server search.
    *
    * @param    ds
    *           A bound directory service instance.  If this is 
    *           <code>null</code> then the instance will be dynamically
    *           queried and bound for the duration of the method and then
    *           unbound at the end.  This is a convenience for situations
    *           in which only a single value is needed.  This has negative
    *           performance implications in the case where multiple values
    *           are sequentially obtained.
    * @param    nodeId
    *           ID of node to obtain.  This will be appended to the various
    *           paths listed above (and in the precedence order specified 
    *           above).
    * @param    initial
    *           The default value to return if the directory backed value
    *           cannot be obtained for ANY reason.
    * @param    scope 
    *           Search mode for the directory node lookup.
    * @param    accountIds
    *           Explicit list of account IDs which should be included in the search. When
    *           {@code null}, the current context's account IDs are resolved and used.
    *
    * @return   The requested value or the given initial value if for ANY
    *           reason the directory backed value cannot be obtained.
    */
   public static boolean getDirectoryNodeBoolean(DirectoryService ds,
                                                 String           nodeId,
                                                 boolean          initial,
                                                 DirScope         scope,
                                                 String[]         accountIds)
   {
      DirectoryHelper helper = new DirectoryHelper("value")
      {
         @Override
         public Object getDirectoryNode(DirectoryService ds, String nodeId)
         {
            return ds.getNodeBoolean(nodeId, attrName);
         }
      };
      
      Boolean result = (Boolean) getDirectoryNodeWorker(ds,
                                                        nodeId,
                                                        "boolean",
                                                        helper,
                                                        scope,
                                                        accountIds);
      
      return (result == null) ? initial : result.booleanValue();
   }

   /**
    * Obtain the specified attribute from the directory if it exists or return
    * the default value if the directory node doesn't exist with the proper
    * type, the directory paths don't exist or the directory is unavailable. 
    * <p>
    * This method implements a search algorithm that allows values to be
    * found that are account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/&lt;id&gt;
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/&lt;id&gt;
    * <p>
    * No exceptions are thrown, the given initial value will always be safely
    * returned.
    * <p>
    * If the {@code scope} is set to {@link DirScope#BOTH}, first a per-account search will be
    * performed.  If the value is not found, it fallbacks to a per-server search.
    *
    * @param    ds
    *           A bound directory service instance.  If this is 
    *           <code>null</code> then the instance will be dynamically
    *           queried and bound for the duration of the method and then
    *           unbound at the end.  This is a convenience for situations
    *           in which only a single value is needed.  This has negative
    *           performance implications in the case where multiple values
    *           are sequentially obtained.
    * @param    nodeId
    *           ID of node to obtain.  This will be appended to the various
    *           paths listed above (and in the precedence order specified 
    *           above).
    * @param    initial
    *           The default value to return if the directory backed value
    *           cannot be obtained for ANY reason.
    * @param    scope 
    *           Search mode for the directory node lookup.
    * @param    accountIds
    *           Explicit list of account IDs which should be included in the search. When
    *           {@code null}, the current context's account IDs are resolved and used.
    *
    * @return   The requested value or the given initial value if for ANY
    *           reason the directory backed value cannot be obtained.
    */
   public static byte[] getDirectoryNodeByteArray(DirectoryService ds,
                                                  String           nodeId,
                                                  byte[]           initial,
                                                  DirScope         scope,
                                                  String[]         accountIds)
   {
      DirectoryHelper helper = new DirectoryHelper("value")
      {
         @Override
         public Object getDirectoryNode(DirectoryService ds, String nodeId)
         {
            return ds.getNodeByteArray(nodeId, attrName);
         }
      };
      
      byte[] result = (byte[]) getDirectoryNodeWorker(ds,
                                                      nodeId,
                                                      "bytes",
                                                      helper,
                                                      scope,
                                                      accountIds);
      
      return (result == null) ? initial : result;
   }

   /**
    * Obtain the path to the node in the directory if it exists or return
    * <code>null</code> if the directory node doesn't exist with the proper
    * type, the directory paths don't exist or the directory is unavailable. 
    * <p>
    * This method implements a search algorithm that allows paths to be
    * found that are account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * There are two search scopes: account relayed and server related.
    * Account related search is described below. The server related search
    * is a version of the search that is limited to the first two server
    * related levels.
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/&lt;id&gt;
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/&lt;id&gt;
    * <p>
    * If the {@code scope} is set to {@link DirScope#BOTH}, first a per-account search will be
    * performed.  If the value is not found, it fallbacks to a per-server search.
    *
    * @param    ds
    *           A bound directory service instance.  If this is 
    *           <code>null</code> then the instance will be dynamically
    *           queried and bound for the duration of the method and then
    *           unbound at the end.  This is a convenience for situations
    *           in which only a single value is needed.  This has negative
    *           performance implications in the case where multiple values
    *           are sequentially obtained.
    * @param    nodeId
    *           ID of node to obtain.  This will be appended to the various
    *           paths listed above (and in the precedence order specified 
    *           above).
    * @param    expected
    *           Expected class of node.  This will be checked with that
    *           found in the directory for consistency.
    * @param    scope 
    *           Search mode for the directory node lookup.
    * @param    accountIds
    *           Explicit list of account IDs which should be included in the search. When
    *           {@code null}, the current context's account IDs are resolved and used.
    *
    * @return   The path to the found node or <code>null</code> if for ANY
    *           reason the directory node cannot be found.
    */
   public static String findDirectoryNodePath(DirectoryService ds,
                                              String           nodeId,
                                              String           expected,
                                              DirScope         scope,
                                              String[]         accountIds)
   {
      DirectoryHelper helper = new DirectoryHelper("")
      {
         /**
          * @param ds  
          */
         @Override
         public Object getDirectoryNode(DirectoryService ds, String nodeId)
         {
            // simply returns the path when called, this works because the
            // calling code has already checked the path for existence and
            // type matches
            return nodeId;
         }
      };
      
      return (String) getDirectoryNodeWorker(ds, nodeId, expected, helper,
                                             scope, accountIds);
   }
   
   /**
    * Obtain the specified attribute from the directory if it exists or return
    * the <code>null</code> if the directory node doesn't exist with the
    * proper type, the directory paths don't exist or the directory is 
    * unavailable. 
    * <p>
    * This method implements a search algorithm that allows values to be
    * found that are account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/&lt;id&gt;
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/&lt;id&gt;
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/&lt;id&gt;
    * <p>
    * No exceptions are thrown, the requested attribute of <code>null</code>
    * will always be safely returned.
    * <p>
    * If the {@code scope} is set to {@link DirScope#BOTH}, first a per-account search will be
    * performed.  If the value is not found, it fallbacks to a per-server search.
    * 
    * @param    ds
    *           A bound directory service instance.  If this is 
    *           <code>null</code> then the instance will be dynamically
    *           queried and bound for the duration of the method and then
    *           unbound at the end.  This is a convenience for situations
    *           in which only a single value is needed.  This has negative
    *           performance implications in the case where multiple values
    *           are sequentially obtained.
    * @param    nodeId
    *           ID of node to obtain.  This will be appended to the various
    *           paths listed above (and in the precedence order specified 
    *           above).
    * @param    expected
    *           Expected class of node or <code>null</code>. If given, this  
    *           will be checked with that found in the directory for
    *           consistency.
    * @param    helper
    *           The data type-specific helper to query the given node.
    * @param    scope 
    *           Search mode for the directory node lookup.
    * @param    accountIds
    *           Explicit list of account IDs which should be included in the search. When
    *           {@code null}, the current context's account IDs are resolved and used.
    *
    * @return   The requested value or <code>null</code> if for ANY
    *           reason the directory backed value cannot be obtained.
    */
   @SuppressWarnings("null")
   public static Object getDirectoryNodeWorker(DirectoryService ds,
                                               String           nodeId,
                                               String           expected,
                                               DirectoryHelper  helper,
                                               DirScope         scope,
                                               String[]         accountIds)
   {
      boolean needsBind = (ds == null);
      
      // implement "one-shot" mode as a convenience
      if (needsBind)
      {
         // access the directory service
         ds = DirectoryService.getInstance();
         
         // is the directory service is unavailable?
         if (ds == null)
         {
            // done!
            return null;
         }
         
         needsBind = !ds.isBound();
         
         // try to bind
         if (needsBind && !ds.bind())
         {
            // can't access the directory
            return null;
         }
      }
      
      // our result
      Object result = null;
      
      // any failure from here out, triggers an unbind if needed
      try
      {
         // access the security manager
         SecurityManager sm = SecurityManager.getInstance();
         
         // any failure?
         if (sm == null && scope == DirScope.ACCOUNTS)
         {
            // can't access the security manager
            return null;
         }
         
         // obtain our context-specific values
         String   serverId   = ds.getServerId();
         String[] nullAcct = new String[] { null };
         if (accountIds == null && (scope == DirScope.BOTH || scope == DirScope.ACCOUNTS))
         {
            accountIds = sm.getAccountIds();
         }
   
         // check account availability
         if (scope == DirScope.ACCOUNTS && accountIds.length == 0)
         {
            // something is wrong with our context
            return null;
         }
   
         // our search algorithm's precedence order
         // 1. /server/<serverID>/runtime/<account_or_group>/<id>
         // 2. /server/<serverID>/runtime/default/<id>
         // 3. /server/default/runtime/<account_or_group>/<id>
         // 4. /server/default/runtime/default/<id>
         // or
         // 1. /server/<serverID>/<id>
         // 2. /server/default/<id>
         
         // if search scope is "BOTH", the precedence order is per-account then per-server.
         
         StringBuffer clsb = new StringBuffer("/meta/class/");
         String cls = expected == null ? null : 
                                         clsb.append(expected).toString();
         
         String pathSrv = "/server/" + serverId;
         
         // check the specific server first and then the global default
         String[] rootPaths = null;
         if (scope == DirScope.ACCOUNTS)
         {
            rootPaths = new String[]
                        {
                           pathSrv + "/runtime",
                           "/server/default/runtime"
                        };
         }
         else if (scope == DirScope.SERVER)
         {
            rootPaths = new String[]
                        {
                           pathSrv,
                           "/server/default"
                        };
         }
         else
         {
            // use a fallback approach: first by account and last by server
            rootPaths = new String[]
                        {
                           pathSrv + "/runtime",
                           "/server/default/runtime",
                           pathSrv,
                           "/server/default"
                        };
         }
         
         // loop through the target paths
         outer:
         for (int i = 0; i < rootPaths.length; i++)
         {
            boolean onSrvPath = (scope == DirScope.SERVER || (scope == DirScope.BOTH && i >= 2));
            
            // when searching per-server, use the "null" as account
            // when searching per-account, use the real accounts
            String[] accIds = (onSrvPath ? nullAcct : accountIds);
            
            // check each account
            for (int j = 0; j < accIds.length; j++)
            {
               result = executeLookup(ds,
                                      rootPaths[i],
                                      accIds[j],
                                      nodeId,
                                      cls,
                                      helper);
               if (result != null)
               {
                  break outer;
               }
            }
            
            if (onSrvPath)
               continue;
            
            // check for a default for all accounts
            result = executeLookup(ds,
                                   rootPaths[i],
                                   "default",
                                   nodeId,
                                   cls,
                                   helper);
            if (result != null)
            {
               break outer;
            }
         }         
      }
      
      catch (Exception exc)
      {
         // just allow null to be returned
      }

      finally
      {
         // if we are one-shot, cleanup
         if (needsBind)
         {
            // last use of the directory, end our session with it
            ds.unbind();
         }
      }
      
      return result;
   }

   /**
    * Removes the specified directory node, including all child nodes.
    *
    * @param    ds
    *           A bound directory service instance with an editing batch open 
    * @param    nodeId
    *           ID of node to remove
    *
    * @return   <code>true</code> if removed the node successfully
    */
   public static boolean removeDirectoryNode(DirectoryService ds,
                                             String           nodeId)
   {
      // enumerate and remove all children first, recursively
      String[] child = ds.enumerateNodes(nodeId);
      if (child == null)
      {
         return false;
      }
      
      for (int i = 0; i < child.length; i++)
      {
         if (!removeDirectoryNode(ds, nodeId + "/" + child[i]))
         {
            return false;
         }
      }
      
      // remove the specified node now 
      return ds.deleteNode(nodeId);
   }
   
   /**
    * Obtain the port number which has been mapped in the directory to a given
    * service name.  This is a partial replacement for the service normally
    * provided by Unix' <code>getservbyname()</code> API.
    * <p>
    * It is <i>partial</i> because it only will return a valid port number for
    * services explicitly mapped in the directory.  The current implementation
    * knows nothing of the mappings provided in the standard services file.
    * Thus, it should be used only for custom services mapped to ports in the
    * directory.
    * <p>
    * The following directory paths are checked for mappings in the following
    * precedence order:
    * <p>
    * <code>/server/&lt;server_id&gt;/port_services/&lt;service&gt;</code><br>
    * <code>/server/default/port_services/&lt;service&gt;</code>
    * <p>
    * where:
    * <ul>
    *   <li><code>&lt;server_id&gt;</code> is the unique identifier of the
    *       current P2J server instance;  and</li>.
    *   <li><code>&lt;service&gt;</code> is the service name given by the
    *       <code>name</code> parameter to this method.</li>.
    * </ul>
    *
    * @param   ds
    *          A bound directory service instance.  If this is 
    *          <code>null</code> then the instance will be dynamically
    *          queried and bound for the duration of the method and then
    *          unbound at the end.  This is a convenience for situations
    *          in which only a single value is needed.  This has negative
    *          performance implications in the case where multiple values
    *          are sequentially obtained.
    * @param   name
    *          Service name to look up.
    *
    * @return  The requested value or -1 if the given service is not mapped to
    *          any port number in the directory paths listed above.
    */
   public static int getPortForService(DirectoryService ds, String name)
   {
      DirectoryHelper helper = new DirectoryHelper("value")
      {
         @Override
         public Object getDirectoryNode(DirectoryService ds, String nodeId)
         {
            return ds.getNodeInteger(nodeId, attrName);
         }
      };
      
      StringBuilder buf = new StringBuilder("port_services/");
      buf.append(name);
      Integer result = (Integer) getDirectoryNodeWorker(ds,
                                                        buf.toString(),
                                                        "integer",
                                                        helper,
                                                        DirScope.SERVER,
                                                        null);
      
      return ((result == null) ? -1 : result.intValue());
   }
      
   /**
    * Serves as a placeholder for some interactive help.
    */
   public static void noHelp()
   {
      LogicalTerminal.messageBox("No application help is available");
   }
   
   /**
    * Get the short name of a class or interface (everything after the last
    * dot in the fully qualified class name).
    *
    * @param   cls
    *          Class.
    *
    * @return  Unqualified class name.
    */
   public static String getUnqualifiedName(Class<?> cls)
   {
      String name = cls.getName();
      
      return name.substring(name.lastIndexOf(".") + 1);
   }
   
   /**
    * Find as many files as case-insensitively match the file specified by
    * <code>fullPath</code>, where the directory specified by
    * <code>leading</code> represents a leading portion of the given path.
    * <p>
    * This method calls itself recursively with progressively more specific
    * <code>leading</code> directories, until <code>leading</code> represents
    * the full path qualifier for the file which is the target of the match.
    * <p>
    * All paths in this method are processed using the current system's file
    * separator character.  This means that any input to this method must
    * have already been properly converted to the system file separator.
    * 
    * @param   leading
    *          A leading portion of the directory qualifier for the
    *          <code>fullPath</code> argument. The top-level call to this
    *          method would pass <code>null</code> here.  Recursive calls will
    *          pass a value. If given, this MUST be a valid directory in the
    *          file system AND it must be a valid prefix for the target file
    *          (it must case-insensitively match the begining of the target 
    *          filename).
    * @param   fullPath
    *          A canonicalized path name which represents an absolute path of
    *          the file to be matched.  Must NOT be <code>null</code>.
    * @param   list
    *          A list in which to store matching <code>File</code> objects
    *          (which is the result of this processing).  Must NOT be
    *          <code>null</code>.
    */
   private static void matchPathCaseInsensitively(File leading,
                                                  String fullPath,
                                                  List<File> list)
   {
      final String SEP = File.separator;
      int pos = 0;
      
      // process leading portion (pull out the directory name and calculate
      // its length when ending with the file separator character)
      if (leading != null)
      {
         String dirName = leading.getAbsolutePath();
         pos = dirName.length();
         
         // safety check
         if (!dirName.equalsIgnoreCase(fullPath.substring(0, pos)))
         {
            // TODO: add error reporting/logging
            
            // leading portion cannot match the target file name
            return;
         }
         
         if (!dirName.endsWith(SEP))
         {
            pos += SEP.length(); // always 1, from File.separator javadoc
         }
      }
      
      // gather everything after the leading portion
      String remaining = fullPath.substring(pos);
      
      // find the end of the next path segment; this is the last segment (which
      // is the base file name itself) if this is -1
      pos = remaining.indexOf(SEP);
      
      // extract the portion of the path which corresponds with the directory
      // level we're currently scanning; this is a path segment only
      final String match = (pos >= 0
                            ? remaining.substring(0, pos)
                            : remaining);
      
      // file filter which matches case-insensitively on the match string
      final FileFilter ff = new FileFilter()
      {
         public boolean accept(File test)
         {
            // this only checks one segment at a time, not portions that
            // "cross" or include file separator characters
            String name = test.getName();
            if (name.isEmpty())
            {
               // special care on windows: when comparing the drive (C:\) segments the name
               // is empty because there is no directory.
               // manually comparing the drives:
               return test.getAbsolutePath().equalsIgnoreCase(match + SEP);
            }
            
            return name.equalsIgnoreCase(match);
         }
      };
      
      File[] files = null;
      
      if (leading == null)
      {
         // top-level call to this method
         
         // special case:  no leading directory specified, so we start at the
         // file system roots and try to match the portion of our path up to
         // the first file separator
         File[] roots = File.listRoots();
         List<File> l = new ArrayList<>(roots.length);
         for (File next : roots)
         {
            if (ff.accept(next))
            {
               l.add(next);
            }
         }
         
         files = l.toArray(new File[l.size()]);
      }
      else
      {
         // recursive call to this method
         
         // safety code; leading must always be a directory
         if (!leading.isDirectory())
         {
            // TODO: add error reporting/logging
            
            // this is definitely not a match
            return;
         }
         
         files = leading.listFiles(ff);
         
         if (files == null)
         {
            // TODO: add error reporting/logging
            
            // this should only occur on an I/O error
            return;
         }
      }
      
      if (pos < 0)
      {
         // last level of search; the contents of the files array contains our final result(s)
          Collections.addAll(list, files);
      }
      else
      {
         // search recursively down through next set of matching directories
         for (File next : files)
         {
            // file matches cannot be checked further using recursion and they
            // can't match anyway, so avoid them
            if (next.isDirectory())
            {
                matchPathCaseInsensitively(next, fullPath, list);
            }
         }
      }
   }
   
   /**
    * Builds the node path from its parts, checks that it exists with the
    * correct data type and if so, returns the value.  The node path that
    * will be searched is:
    * <p>
    * <pre>
    * root + "/" + postfix + "/" + nodeId 
    * </pre>
    *
    * @param    ds
    *           A bound directory service instance.  Must not be 
    *           <code>null</code>.
    * @param    root
    *           First part of the path to build.
    * @param    postfix
    *           Second part of the path to build or <code>null</code>.
    * @param    nodeId
    *           ID of node to obtain.  This will be appended to the two
    *           paths listed above.
    * @param    expected
    *           Expected class of node or <code>null</code>. If given, this  
    *           will be checked with that found in the directory for
    *           consistency.
    * @param    helper
    *           The data type-specific helper to query the given node.
    *
    * @return   The requested value or <code>null</code> if for ANY
    *           reason the directory backed value cannot be obtained.
    */
   private static Object executeLookup(DirectoryService ds,
                                       String           root,
                                       String           postfix,
                                       String           nodeId,
                                       String           expected,
                                       DirectoryHelper  helper)
   {
      StringBuffer path = new StringBuffer();      
      path.append(root).append("/");
      if (postfix != null)
         path.append(postfix).append("/");
      path.append(nodeId);
      
      String current   = path.toString();
      String nodeClass = ds.getNodeClass(current);
      
      if (nodeClass != null)
      {
         if (expected == null || nodeClass.equals(expected))
         {
            // we found our path!
            return helper.getDirectoryNode(ds, current);
         }
         else
         {
            // TODO: log the fact that a node existed but did not match the
            //       proper datattype
         }
      }
      
      // nothing found or data type is incorrect, continue search
      return null;
   }

   /**
    * Generic type safe equals. It can be used for <code>null</code> objects
    * without fear of NPE. If both provided objects are <code>null</code> then
    * they are assumed equal.
    * @param   <T>
    *          Type parameter.
    * @param   obj1
    *          First object for comparison.
    * @param   obj2
    *          Second object for comparison.
    *
    * @return  <code>true</code> if objects are equal.
    */
   public static<T> boolean equals(T obj1, T obj2)
   {
      if (obj1 == null && obj2 == null)
         return true;

      return (obj1 != null) ? obj1.equals(obj2) : obj2.equals(obj1);
   }

   /**
    * Provides a simple unified API for delegating the querying of a
    * data type-specific value from the directory.
    */
   private static abstract class DirectoryHelper
   {
      /** Attribute name to obtain. */
      protected String attrName = null;
      
      /**
       * Construct an instance and save the attribute name for future use.
       *
       * @param    attrName
       *           The name of the attribute to obtain.
       */
      public DirectoryHelper(String attrName)
      {
         this.attrName = attrName;
      }
      
      /** 
       * User-implemented method which queries the given directory service
       * for the attribute value of the given node ID and the returns the 
       * result.  The implementation of this method if data type-specific
       * however the result is returned generically.
       * <p>
       * The attribute name provided at construction time is the one to use
       * during the query.  This eliminates the need to pass the attribute
       * name to the code to which the call is delegated.
       *
       * @param    ds
       *           A bound directory service instance.  Must not be 
       *           <code>null</code>.
       * @param    nodeId
       *           The full path of the node to obtain.
       *           
       * @return   Attribute value. 
       */
      public abstract Object getDirectoryNode(DirectoryService ds,
                                              String           nodeId);
   }

   /**
    * Returns a context-wide unique integer value which can be used as an ID.
    * If the internal counter reaches the maximum value, it wraps around to 0.
    * <p>
    * @return   next available integer value
    */
   public static int uniqueId()
   {
      WorkArea wa = work.get();
      
      int newValue = 0;
      
      synchronized (wa)
      {
         newValue = ++wa.counter;
         if (newValue == Integer.MAX_VALUE)
            wa.counter = -1;
      }
      
      return newValue;
   }      
   
   /**
    * Check if the two sets are the same. Note that <code>null</code> instances are allowed for 
    * either set.
    * 
    * @param    a
    *           The first set.
    * @param    b
    *           The second set.
    * 
    * @return   <code>true</code> if the sets match.
    */
   public static<T> boolean isSame(Set<T> a, Set<T> b)
   {
      if ((a == null && b != null) || (a != null && b == null))
         return false;
         
      if (a == null && b == null)
         return true;
      
      return a.size() == b.size() && a.containsAll(b);
   }
   
   /**
    * Determines whether two sets have common elements.
    * 
    * @param <T> 
    *
    * @param    a
    *           a set
    * @param    b
    *           another set
    *
    * @return   <code>true</code> if there are common elements
    */
   public static<T> boolean intersect(Set<T> a, Set<T> b)
   {
      Iterator<T> iter = a.iterator();
      Object o = null;
      
      while (iter.hasNext())
      {
         o = iter.next();
         if (b.contains(o))
            return true;
      }
      
      return false;
   }
   
   /**
    * Produces the intersection of two sets and puts it into the third set.
    * 
    * @param <T> 
    *
    * @param    a
    *           a set
    * @param    b
    *           another set
    * @param    result
    *           result set
    */
   public static<T> void intersection(Set<T> a, Set<T> b, Set<T> result)
   {
      result.clear();
      
      Iterator<T> iter = a.iterator();
      T o = null;
      
      while (iter.hasNext())
      {
         o = iter.next();
         if (b.contains(o))
            result.add(o);
      }
   }
   
   /**
    * Produces the difference of two sets and puts it into the third set.
    * @param <T> 
    *
    * @param    a
    *           a set
    * @param    b
    *           another set, which is to be subtracted from 'a'
    * @param    result
    *           result set
    */
   public static<T> void difference(Set<T> a, Set<T> b, Set<T> result)
   {
      result.clear();
      
      Iterator<T> iter = a.iterator();
      T o = null;
      
      while (iter.hasNext())
      {
         o = iter.next();
         if (!b.contains(o))
            result.add(o);
      }
   }
   
   /**
    * Produces the difference of two int arrays.
    *
    * @param    a
    *           The first array.
    * @param    b
    *           The second array which is to be subtracted from the first.
    *
    * @return   All elements from the first array that are not in the
    *           second array.
    */
   public static int[] diffIntArrays(int[] a, int[] b)
   {
      // boundary cases
      if (a.length == 0)
      {
         return new int[] {};
      }

      int[] aa = new int[a.length];
      for (int i = 0; i < a.length; i ++)
      {
         aa[i] = a[i];
      }
         
      Arrays.sort(aa);

      if (b.length == 0)
      {
         return aa;
      }
      
      int[] bb = new int[b.length];
      for (int i = 0; i < b.length; i ++)
      {
         bb[i] = b[i];
      }
         
      Arrays.sort(bb);

      int[] res = new int[aa.length];
      int count = 0;
      int i = 0;
      int j = 0;
      
      scan:
      for (i = 0; i < aa.length; i ++)
      {
         // equal numbers; skip both
         if (aa[i] == bb[j])
         {
            j ++;
            if (j < bb.length)
            {
               continue;
            }
            
            break scan;
         }
         
         // integer from a to keep
         if (a[i] < b[j])
         {
            res[count ++] = a[i];
            continue;
         }
         
         // integer from b to skip
         while (a[i] > b[j])
         {
            j ++;
            if (j < bb.length)
            {
               continue;
            }
            
            break scan;
         }
      }

      // copy the rest of a
      for (int k = i + 1; k < aa.length; k ++)
      {
         res[count ++] = aa[k];
      }
            
      // compact the array
      int[] diff = new int[count];
      for (i = 0; i < count; i ++)
      {
         diff[i] = res[i];
      }
      
      return diff;
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** Internal counter tha last used value */ 
      private int counter = -1; 
   }
   
   /**
    * Simple container that stores and returns a context-local instance of
    * the global work area.
    */
   private static class ContextContainer
   extends ContextLocal<WorkArea>
   {
      /**
       * Initializes the work area, the first time it is requested within a
       * new context.
       *
       * @return   The newly instantiated work area.
       */
      @Override
      protected WorkArea initialValue()
      {
         WorkArea wa = new WorkArea();
         return wa;
      }   
   }
}
